#
# Global Cache GC-IRL and similar - IR receiver interface
#
# Copyright 2010-2016 Perry The Cynic. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import asyn
import ir
import gcir
import re
import sys
import time

BAUDRATE=9600


class Error(Exception):
	""" GC-IRL originated error. There's only one, and its name is "unknowncommand". :) """


#
# A GC Infrared listener device (GC-IRL or similar)
#
class IRL(asyn.Callable, ir.Receiver):
	""" a GC-IRL, GC-IRE, or similar device, with its primitive command set.

		GC-IRL and GC-IRE are serial-attached plug-size devices that read
		Consumer IR signaling and communicate to the serial line.
		They have a primitive but serviceable command set to deal with
		transmission parameters and error conditions.

		Note that an IRL is not a Selectable. You hand it a Selectable
		representing the serial port, which it takes over completely.

		The learning mode of an iTach is functionally similarly but is not modeled here.
		It is handled directly in GCNet.
	"""
	#
	# Scannable state machine
	#
	scan_idle = asyn.scan.Regex([						# [command state]
		(r'\0?GC-IR[LE],(.*?)\r', "IR batch"),			# complete IR event
		(r'\0?GC-IR[LE],(\d+)(?=,)', "IR"),				# beginning of IR event
		(r'\0?device,(.*?)\r', "device"),				# device reply
		(r'\0?ver,(.*?)\r', "ver"),						# version reply
		(r'\0?IRend,(\d+)msec\r', "IRend"),				# timeout reply
		(r'\0?(.*?)\r', "other")						# unexpected inbound line
	])
	scan_pulse = asyn.scan.Regex([						# [receiving pulsetrain state]
		(r',(\d+,\d+)(?=[^0-9])', "IR pulse"),			# uncompressed pulse
		(r'([A-W])', "IR repeat"),						# compression code
		(r'\r', "IR end"),								# normal end of IR stream
		(r'(.*?)\r', "IR abort")						# unexpected end of IR stream
	])

	def __init__(self, selectable, callout=None):
		""" Create an IR reader based on an asyn.Selectable.

			Works with a serial or network stream.
			This will take ownership of the Selectable and configure
			it at will. Closing the IRL will also close it.
		"""
		asyn.Callable.__init__(self, callout=callout)
		ir.Receiver.__init__(self)
		self.device = None								# result of (id)
		self.version = None								# result of (gv)
		self.timeout = None								# result of e[1-4]
		self._build = None								# IREvent in progress or None
		self._threshold = sys.maxint					# full-signal delivery only

		self._dev = selectable
		self._dev.scan = IRL.scan_idle					# command state
		self._dev.add_callout(self._calldown)

	def close(self):
		self._dev.close()

	def delivery_threshold(self, thr):
		""" Deliver partial events if at least (thr) Pulses have been recorded. Use None for final delivery only. """
		self._threshold = sys.maxint if thr is None else thr

	def command(self, cmd):
		""" Send a command to the IRL. """
		self._dev.write(cmd + '\r')

	def identify(self):
		""" Issue identification commands. """
		self.command('')		# un-wedge the receiver
		self.command('id')		# identify
		self.command('gv')		# get version

	#
	# Unified input callback.
	#
	def _calldown(self, ctx, data=None):
		if ctx.error:
			return self.callout(ctx)
		try:
			if ctx.state == 'device':
				self.device = data
				self.callout('status')
			elif ctx.state == 'ver':
				self.version = data
				self.callout('status')
			elif ctx.state == 'IRend':
				self.timeout = int(data)
				self.callout('timeout', self.timeout)
			elif ctx.state == 'other':
				self.callout(Error('unexpected data', data))
			elif ctx.state == 'IR batch':
				self.callout('event', gcir.Event(data)) # build and deliver immediately
			elif ctx.state == 'IR':
				self._build = gcir.Event()				# incremental build
				self._build.set_carrier(data)			# set carrier
				self._dev.scan = IRL.scan_pulse			# start collecting pulses
				self._build_start = time.time()			# wall clock timer
			elif ctx.state == "IR pulse":
				self._build.add_pulse(data)
			elif ctx.state == 'IR repeat':
				self._build.add_repeat(data)
			elif ctx.state == 'IR end':
				self._build.finish()					# finish incremental build
				self.callout('event', self._build)		# deliver IR object
				self._build = None						# clear buffer
				self._dev.scan = IRL.scan_idle			# back to base state
				del self._build_start
			elif ctx.state == 'IR abort':
				self._build = None						# discard IR object
				self._dev.scan = IRL.scan_idle			# back to base state
				del self._build_start
				self.callout(Error("IR Receiver abort", data))
		except gcir.Error, e:							# Event build threw - reset
			self._build = None

		# perform partial delivery if requested
		if self._build is not None:
			if len(self._build.pulsetrain) >= self._threshold:
				assert not self._build.is_complete()
				self.callout('incoming', self._build) # deliver and keep partial signal


	def __repr__(self):
		r = "<%s" % (self.device or "GC-IRx")
		if self.version: r += " %s" % self.version
		if self.timeout: r += " %dmsec" % self.timeout
		return r + ("@%s>" % hex(id(self)))


#
# Regression test shell
#
if __name__ == "__main__":
	arg = sys.argv[1] if len(sys.argv) > 1 else '/dev/cu.usbserial'

	def commands(ctx, cmd=None):
		if ctx.state == 'END':
			return control.close()
		elif ctx.state == 'command':
			irl.command(cmd)

	def irl_event(ctx, arg=None):
		if ctx.state == 'status':
			if irl.device: print "device=%s" % irl.device
			if irl.version: print "version=%s" % irl.version
		elif ctx.state == 'timeout':
			print "Timeout", arg
		elif ctx.state == 'incoming':
			pass
		elif ctx.state == 'event':
			if not arg.is_complete():
				sys.stdout.write("                  \rIn: %d %g" % (
					len(arg.pulsetrain), time.time() - arg.inception))
				sys.stdout.flush()
			else:
				print "\nReceived:", repr(arg)
		else:
			print "UNRECOGNIZED:", ctx, arg

	control = asyn.Controller()
	control.commands(commands)
	if '/' in arg:		# path to serial device
		import serial
		seq = serial.Serial(arg, baudrate=BAUDRATE)
	else:				# hostname, assume first serial port of GC-100-alike
		import socket
		seq = socket.create_connection((arg, 4999))
	irl = IRL(control.file(seq), callout=irl_event)
	irl.delivery_threshold(2)
	print "GC-IRL regression test - interactive mode. Enter command or send IR. ^D to exit."
	control.run()
