#
# jvc - JVC CIR signals
#
# This coding scheme comes from old JVC VCRs. I have no idea how
# widespread, or how common, it really is.
#
# Copyright 2011-2016 Perry The Cynic. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import ir

name = "JVC"
exactMatch = True


#
# Characteristics of the JVC protocol.
#
# This is very similar to the NEC "protocol" - the lead-in/zero/one
# codes are close enough to be an aberration. But instead of having
# a distinct repeat trailer, this code simply repeats everything
# except for the lead-in; the trailer is part of each repeat group.
#
CARRIER = 38000
LEADER = ir.Pulse(320,158)
ZERO = ir.Pulse(21,19)
ONE = ir.Pulse(21,58)
TRAILER = ir.Pulse(21,860)
TRAILER_SLOP = 200


class Symbol(ir.Symbol):

	carrier = CARRIER

	def __init__(self, dev, func):
		ir.Symbol.__init__(self, module)
		self.device = dev
		self.function = func

	def matches(self, code):
		return (isinstance(code, Code) and
			self.device == code.device and
			self.function == code.function)

	def code(self, repeats=0):
		return Code(self.device, self.function, repeats)

	#
	# Canonical string form
	#
	def __str__(self):
		return "<%s:%02x/%02x>" % (self.device, name, self.function)

	def __repr__(self): return str(self)

	@staticmethod
	def from_text(text):
		dev, fun = text.split('/')
		return Symbol(int(dev, 8), int(fun, 8))


class Code(ir.Code):

	carrier = CARRIER

	def __init__(self, dev, func, repeat):
		ir.Code.__init__(self, module)
		self.device = dev
		self.function = func
		self.repeat = repeat

	def __str__(self):
		return "<Code:JVC %02x/%02x{%d}>" % (self.device, self.function, self.repeat)

	def __repr__(self): return str(self)

	def symbol(self):
		return Symbol(self.device, self.function)

	def form(self):
		""" Make abstract IR form. Returns (base, repeat, count) tuple. """
		func = self.function
		def bits(mask, count):
			for n in range(0, count): yield mask & 1; mask = mask >> 1
		run = ([LEADER]
		+	[ONE if x else ZERO for x in bits(func, 16)]
		+	[TRAILER])
		if self.repeat:
			return (base, base, self.repeat-1)
		else:
			return (base, None, 0)


#
# Type check and encode
#
def claim(event):
	""" If event looks like a JVC signal, encode it as our Code. Return None if it doesn't. """

	#
	# A JVC pulsetrain looks like this:
	#	LEADER ( ( ZERO | ONE ){16} TRAILER ) *
	#
	try:
		if event.carrier != CARRIER: return None
		if event[0] != LEADER: return None
		if not TRAILER.__eq__(event[17], TRAILER_SLOP): return None

		pulses = event.pulsetrain

		# count and check repeat patterns
		repeats = (len(pulses) - 1) / 17
		if pulses[18:] != pulses[1:18] * (repeats - 1): return None	# check repetitions

		# extract bits
		def bit(p):
			if p == ZERO: return 0
			elif p == ONE: return 1
			else: raise ValueError("invalid bit")
		code = map(bit, pulses[1:17])
		code.reverse() # low-endian

		# Code is now: [func:8][dev:8]
		def bits(slice): return reduce(lambda a, b: a*2+b, slice, 0)
		dev = bits(code[8:16])
		func = bits(code[0:8])
		return Code(dev, func, repeats)

	except (ValueError, IndexError): return None


#
# Fetch our own module reference
#
import sys
module = sys.modules[__name__]
